from __future__ import annotations
from typing import Iterable, NamedTuple, Callable, Any, \
    TypeVar, Generic, NoReturn, ParamSpec, Type, Literal, \
    Sequence, Final
from functools import wraps
from enum import IntEnum, auto
from dataclasses import dataclass

from more_itertools import peekable
from strenum import StrEnum
from unpackable import Unpackable

from .constants import NEW_LINE, NO_ITEMS, NO_RESULT


RepeatTimes = Literal['forever', 'inf', 'infinite', 'loop']
ColumnSeps = Literal['tab', 'newline', 'space']


FOREVER_OPTS: Final[set[str]] = set(RepeatTimes.__args__)
COL_OPTS: Final[dict[str, str]] = {
  'tab': '\t',
  'newline': '\n',
  'space': ' ',
}


T = TypeVar('T')
P = ParamSpec('P')

Decorator = Callable[[Callable[P, T]], Callable[P, T]]
StrParseFunc = Callable[[str, P], str]
StrCheckFunc = Callable[[str], bool | int]
CheckFunc = Callable[P, bool]
QuitFunc = Callable[P, NoReturn]

Chars = Iterable[str]
Arg = str | int | float
Args = Sequence[Arg]


class Peekable(Generic[T], peekable):
  """Generic and typed `peekable` with convenience methods."""
  iterable: Iterable[T]

  @property
  def is_empty(self) -> bool:
    return _is_empty(self)


Strings = Peekable[str] | Iterable[str]
Input = Strings | Args


class InputStrsSep(NamedTuple):
  strings: Strings | None
  sep: str


class StrSep(NamedTuple):
  string: str | None = None
  sep: str = NEW_LINE


class ErrCode(IntEnum):
  """Shell return codes."""
  none: int = -1

  ok: int = 0
  err: int = 1

  no_result: int = 2
  no_handler: int = 3

  found: int = ok
  not_found: int = 4
  bad_input: int = 5

  true: int = ok
  false: int = err

  @classmethod
  def from_bool(cls: Type[ErrCode], other: bool) -> ErrCode:
    return cls.true if other else cls.false

  @property
  def is_err(self) -> bool:
    if self == ErrCode.none:
      return False

    return self != ErrCode.ok

  @property
  def should_quit(self) -> bool:
    return self != ErrCode.none


class CmdState(StrEnum):
  ok: str = auto()
  err: str = auto()
  quit: str = auto()


@dataclass
class Result(Generic[T], Unpackable):
  result: T | None = None
  code: ErrCode = ErrCode.none
  # state: CmdState = CmdState.ok


@dataclass
class Ok(Result[T]):
  code: ErrCode = ErrCode.ok


@dataclass
class Error(Result[T]):
  code: ErrCode = ErrCode.err


@dataclass
class NotFound(Error[T]):
  code: ErrCode = ErrCode.not_found


@dataclass
class BadInput(Error[T]):
  code: ErrCode = ErrCode.bad_input


Item = T | Result[T] | StrSep | ErrCode | int | bool | None
Items = Iterable[Item[T]] | Peekable[Item[T]]

ItemsFunc = Callable[P, Items[T] | Item[T]]
ItemFunc = Callable[P, Result[Item[T] | T | Any]]


NoResult = Result[None](code=ErrCode.no_result)
NoneFound = NotFound[None]()

ErrResult = Error[None]()
IntError = BadInput[int](NO_RESULT)


def _is_empty(it: peekable) -> bool:
  return it.peek(NO_ITEMS) is NO_ITEMS


def _to_peekable(
  func: Callable[P, Iterable[T]]
) -> Callable[P, Peekable[T]]:
  @wraps(func)
  def new_func(*args: P.args, **kwargs: P.kwargs) -> Peekable[T]:
    gen: Iterable[T] = func(*args, **kwargs)
    return Peekable[T](gen)

  return new_func



from __future__ import annotations
from typing import Iterable, cast
import re

from emoji import demojize, emoji_count, emojize
from nth_py.nth import exclude_lines, gen_lines
from unidecode import unidecode

from ..core.constants import EMPTY_STR, NEW_LINE, SAME_LINE, WHITESPACE_RE
from ..core.decorators import _wrap_check_exit, _wrap_parse_print
from ..core.input import _get_stdin, _get_strings_sep
from ..core.process import _output_items
from ..core.types import Args, Chars, ErrResult, Items, NoResult, Peekable, \
  StrSep, _to_peekable


to_ascii = _wrap_parse_print(unidecode)
to_emoji = _wrap_parse_print(emojize)
from_emoji = _wrap_parse_print(demojize)
has_emoji = _wrap_check_exit(emoji_count)


@_output_items
def nth(*line_nums: int, exclude: bool = False) -> Items[StrSep]:
  """
  Print lines on `line_nums` from standard input.

  Setting the `exclude` flag will instead print all lines from standard input and lines `line_nums`
  will be excluded.
  """
  stdin = _get_stdin()

  if not (stdin and line_nums):
    yield ErrResult
    return

  line_nums: list[int]  # keep mypy quiet
  lines: Iterable[str] | Peekable[str]

  if exclude:
    lines = exclude_lines(line_nums, stdin)

  else:
    lines = gen_lines(line_nums, stdin)

  lines = Peekable[str](lines)

  if lines.is_empty:
    yield NoResult
    return

  yield from map(StrSep, lines)


@_output_items
def col(
  num: int,
  *args: Args,
  sep: str = WHITESPACE_RE,
) -> Items[StrSep]:
  """
  Return the string in column specified by `num`.

  Set `sep` to change the column separator from the whitespace regex default.
  """
  strings, _ = _get_strings_sep(args, strip=False)

  index: int = num - 1
  no_result: bool = True

  for string in strings:
    results = [r for r in re.split(sep, string) if r]

    if len(results) >= num:
      yield StrSep(results[index], NEW_LINE)
      no_result = False

  if no_result:
    yield NoResult


@_output_items
def sbob(
  *args: Args,
  reverse: bool = False,
) -> Items[StrSep]:
  """tYpE lIkE tHiS"""
  strings, sep = _get_strings_sep(args)

  for string in strings:
    chars = _gen_sbob_chars(string, reverse)

    for char in chars:
      yield StrSep(char, SAME_LINE)

    yield StrSep(EMPTY_STR, sep)


@_to_peekable
def _gen_sbob_chars(chars: Chars, reverse: bool = False) -> Chars:
  caps: bool = reverse

  for char in chars:
    if not char.isalpha():
      yield char
      continue

    char = char.upper() if caps else char.lower()
    yield char

    caps = not caps


__all__ = [
  'col',
  'from_emoji',
  'has_emoji',
  'nth',
  'sbob',
  'to_ascii',
  'to_emoji',
]



from .commands import *
from .plugins import *
from .core import NAME


del analyze, default, more, validate, string, transform


def _main():
  from fire import Fire as _Fire

  len = length

  if __name__ == '__main__':
    _Fire(name=NAME)

  else:
    _Fire()


if __name__ == "__main__":
  _main()



from __future__ import annotations

from ...core.decorators import _wrap_check_exit, _use_metadata
from ...core.input import _get_strings_sep
from ...core.process import _output_items
from ...core.types import Args


isalnum = _wrap_check_exit(str.isalnum)
isalpha = _wrap_check_exit(str.isalpha)
isascii = _wrap_check_exit(str.isascii)
isdecimal = _wrap_check_exit(str.isdecimal)
isdigit = _wrap_check_exit(str.isdigit)
isidentifier = _wrap_check_exit(str.isidentifier)
islower = _wrap_check_exit(str.islower)
isnumeric = _wrap_check_exit(str.isnumeric)
isprintable = _wrap_check_exit(str.isprintable)
isspace = _wrap_check_exit(str.isspace)
istitle = _wrap_check_exit(str.istitle)
isupper = _wrap_check_exit(str.isupper)


@_output_items
@_use_metadata(str.endswith)
def endswith(
  suffix: str | tuple[str, ...],
  *args: Args,
  start: int | None = None,
  end: int | None = None,
) -> bool:
  strings, _ = _get_strings_sep(args)
  string: str | None = None

  for string in strings:
    pass

  if string is not None:
    return string.endswith(suffix, start, end)

  return False


@_output_items
@_use_metadata(str.startswith)
def startswith(
  prefix: str | tuple[str, ...],
  *args: Args,
  start: int | None = None,
  end: int | None = None,
) -> bool:
  strings, _ = _get_strings_sep(args)

  if strings.is_empty:
    return False

  first = next(strings)

  return first.startswith(prefix, start, end)


__all__ = [
  'endswith',
  'isalnum',
  'isalpha',
  'isascii',
  'isdecimal',
  'isdigit',
  'isidentifier',
  'islower',
  'isnumeric',
  'isprintable',
  'isspace',
  'istitle',
  'isupper',
  'startswith',
]



from __future__ import annotations
from functools import partial, wraps
from typing import Callable
import sys

from .constants import NEW_LINE, DOCTEST
from .input import _get_strings_sep
from .process import _process_item
from .types import Args, ErrCode, P, T, ItemFunc, QuitFunc, StrParseFunc, \
  Strings, StrCheckFunc, CheckFunc, Decorator


def _use_metadata(
  source: Callable | str,
  name: str = True,
) -> Decorator[P, T]:
  is_callable: bool = isinstance(source, Callable)

  if is_callable:
    docs: str = _strip_doctests(source.__doc__)

  else:
    docs = source

  func_name: str | None = None

  if name and is_callable:
    func_name = source.__name__

  def decorator(to_func: Callable[P, T]) -> Callable[P, T]:
    to_func.__doc__ = docs

    if name and func_name:
      to_func.__name__ = func_name

    return to_func

  return decorator


def _wrap_str_check(func: StrCheckFunc) -> CheckFunc[P]:
  @_use_metadata(func)
  def new_func(*args: Args, **kwargs: P.kwargs) -> bool:
    func_kwargs: CheckFunc[P.args] = partial(func, **kwargs)
    strings, _ = _get_strings_sep(args)

    checks = map(func_kwargs, strings)
    return all(checks)

  return new_func


def _wrap_check_exit(func: StrCheckFunc) -> QuitFunc[str]:
  wrapped = _wrap_str_check(func)
  return _handle_item(wrapped)


def _wrap_parse_print(func: StrParseFunc[P]) -> QuitFunc[P]:
  @_use_metadata(func)
  def new_func(*args: Args):
    strings, sep = _get_strings_sep(args)
    _apply(func, strings, sep)

    sys.exit(ErrCode.ok)

  return new_func


def _handle_item(func: ItemFunc[P, T]) -> QuitFunc[P]:
  @wraps(func)
  def new_func(*args: P.args, **kwargs: P.kwargs):
    result: T = func(*args, **kwargs)
    _process_item(result)

    sys.exit(ErrCode.ok)

  return new_func


def _strip_doctests(string: str | None) -> str | None:
  if string is None:
    return

  if DOCTEST not in string:
    return string

  docs: list[str] = []
  lines = string.split(NEW_LINE)

  for line in lines:
    if DOCTEST in line:
      break

    docs.append(line)

  return NEW_LINE.join(docs)


def _apply(
  func: StrParseFunc[P],
  strings: Strings,
  sep: str,
  *args: P.args,
  **kwargs: P.kwargs,
):
  for string in strings:
    parsed = func(string, *args, **kwargs)
    print(parsed, end=sep)



from __future__ import annotations
from functools import wraps
from typing import Callable
from collections.abc import Iterable
import logging
import sys

from .types import ErrCode, Result, StrSep, QuitFunc, Item, T, \
  ItemsFunc, P, ItemFunc, Items, Peekable


_process_item: QuitFunc[Item[T]]


def _process_item(item: Item[T]):
  match item:
    case Result((string, sep), code):
      if string is not None:
        print(string, end=sep)

      if code.should_quit:
        sys.exit(code)

    case Result(result=bool()) | bool() as result:
      code = ErrCode.from_bool(result)
      sys.exit(code)

    case Result(result, code):
      if result is not None:
        print(result)

      if code.should_quit:
        sys.exit(code)

    case StrSep(string, sep):
      print(string, end=sep)

    case ErrCode() as code if code.should_quit:
      sys.exit(code)

    case None:
      logging.debug(f'No result from handler.')
      sys.exit(ErrCode.no_result)

    case _rest:
      logging.error(f'No handler for {_get_name(type(_rest))}.')
      sys.exit(ErrCode.no_handler)


def _get_name(func: Callable) -> str:
  return func.__name__


def _output_items(func: ItemsFunc[P, T] | ItemFunc[P, T]) -> QuitFunc[P]:
  @wraps(func)
  def new_func(*args: P.args, **kwargs: P.kwargs):
    items: Items[T] | Item[T] = \
      func(*args, **kwargs)

    match items:
      case Result() as item:
        _process_item(item)
        return

      case Iterable():
        items = Peekable[Item[T]](items)

      case _ as item:
        _process_item(item)
        return

    if items.is_empty:
      logging.debug(f'No results found for {_get_name(func)}.')
      sys.exit(ErrCode.no_result)

    for item in items:
      _process_item(item)

    sys.exit(ErrCode.ok)

  return new_func



from __future__ import annotations
from typing import Iterable

from .constants import FIRST, MIN_TIMES, INCREMENT, SKIP, SLICE_SEP
from .types import T


# see https://docs.python.org/3/library/itertools.html#itertools.cycle
def _cycle_times(
  iterable: Iterable[T],
  times: int = MIN_TIMES,
) -> Iterable[T]:
  """Cycle through iterable `times` times."""
  if times < MIN_TIMES:
    return

  saved: list[T] = []

  for element in iterable:
    yield element
    saved.append(element)

  cycles: int = FIRST

  while saved and cycles < times:
    for element in saved:
      yield element

    cycles += INCREMENT


# see https://stackoverflow.com/a/54421070
def _slice_from_str(string: str) -> slice:
  if not string:
    return slice(SKIP)

  indices: list[str | SKIP] = string.split(SLICE_SEP)

  match indices:
    case [end]:
      indices = [SKIP, end]

  nums = (
    int(index) if index else SKIP
    for index in indices
  )

  return slice(*nums)



from __future__ import annotations
from itertools import cycle

from ..core.base import _slice_from_str, _cycle_times
from ..core.constants import NEW_LINE, SAME_LINE, FOREVER
from ..core.input import _get_strings_sep
from ..core.process import _output_items
from ..core.types import Args, Items, StrSep, RepeatTimes, ErrResult, \
  FOREVER_OPTS


_slice = slice


@_output_items
def substring(
  stop: int,
  *args: Args,
  start: int | None = None,
  step: int | None = None,
) -> Items[StrSep]:
  """Return substrings using given indices."""
  strings, _ = _get_strings_sep(args)
  window = _slice(start, stop, step)

  for string in strings:
    sub = string[window]

    if string.endswith(NEW_LINE):
      sep = SAME_LINE

    else:
      sep = NEW_LINE

    yield StrSep(sub, sep)


sub = substring


@_output_items
def slice(
  indices: str,
  *args: Args,
) -> Items[StrSep]:
  """Return substrings using given indices."""
  strings, _ = _get_strings_sep(args)
  window = _slice_from_str(indices)

  for string in strings:
    sub = string[window]

    if string.endswith(NEW_LINE):
      sep = SAME_LINE

    else:
      sep = NEW_LINE

    yield StrSep(sub, sep)


@_output_items
def repeat(
  times: int | RepeatTimes = FOREVER,
  *args: Args,
) -> Items[StrSep]:
  """Repeat string. Set `times` to -1 to repeat forever."""
  match times:
    case str() as s if s.casefold() in FOREVER_OPTS:
      times: int = FOREVER

    case None | 0:
      yield ErrResult
      return

  strings, sep = _get_strings_sep(args)

  if times > FOREVER:
    strings = _cycle_times(strings, times)

  elif times == FOREVER:
    strings = cycle(strings)

  else:
    yield ErrResult
    return

  for string in strings:
    yield StrSep(string, sep)


@_output_items
def contains(
  find: str,
  *args: Args,
  case_sensitive: bool = True
) -> bool:
  """Confirm whether the input contains a given string."""
  find = str(find)

  if not case_sensitive:
    find = find.casefold()

  strings, sep = _get_strings_sep(args)

  for string in strings:
    line = f'{string}{sep}'

    if not case_sensitive:
      line = line.casefold()

    if find in line:
      return True

  return False


__all__ = [
  'contains',
  'slice',
  'substring',
  'repeat',
]



from __future__ import annotations
from typing import Final

import os


NEW_LINE: Final[str] = '\n'
EMPTY_STR: Final[str] = ''
SAME_LINE: Final[str] = EMPTY_STR
SPACE: Final[str] = ' '
DOCTEST: Final[str] = '>>>'
SLICE_SEP: Final[str] = ':'
SH_SEP: Final[str | None] = os.environ.get('IFS')

START_INDEX: Final[int] = 0
MIN_TIMES: Final[int] = 1
FIRST: Final[int] = 1
INCREMENT: Final[int] = 1

NOT_FOUND: int = -1
NO_RESULT: Final[int] = -1
FOREVER: Final[int] = -1
ALL: Final[int] = -1
SKIP: Final[None] = None
NO_ITEMS: Final[Ellipsis] = ...

NO_CMD_ERR: Final[str] = "This command isn't implemented."
WHITESPACE_RE: Final[str] = r'\s'



from __future__ import annotations
from functools import partial
import sys

from .constants import NEW_LINE, EMPTY_STR, SH_SEP
from .types import Args, InputStrsSep, Input, Strings, \
  Arg, _to_peekable


def _is_pipeline() -> bool:
  return not sys.stdin.isatty()


def _get_sep() -> str:
  if SH_SEP is not None:
    return SH_SEP

  return EMPTY_STR if _is_pipeline() else NEW_LINE


def _get_stdin(strip: bool = False) -> Strings | None:
  if not _is_pipeline():
    return None

  parse = partial(_parse_line, strip=strip)
  return map(parse, sys.stdin.buffer)


@_to_peekable
def _get_input(args: Args, strip: bool = False) -> Input:
  if stdin := _get_stdin(strip):
    return stdin

  strings = map(str, args)
  return strings


def _get_strings_sep(args: Args, strip: bool = False) -> InputStrsSep:
  strings = _get_input(args, strip)
  sep = _get_sep()

  return InputStrsSep(strings, sep)


def _parse_line(line: Arg, strip: bool = False) -> str:
  match line:
    case bytes():
      string = line.decode()

    case int() | float():
      string = str(line)

    case str():
      string = line

    case _:
      raise ValueError(f"Not supported: {line}")

  if strip:
    return string.strip()

  return string



from __future__ import annotations
import logging

from ...core.constants import EMPTY_STR, START_INDEX, NOT_FOUND
from ...core.decorators import _use_metadata
from ...core.input import _get_strings_sep
from ...core.process import _output_items
from ...core.types import Args, Result, Ok, IntError, NotFound


@_output_items
def length(*args: Args) -> Result[int]:
  """ Return the length of the string."""
  strings, sep = _get_strings_sep(args)
  sep_size = len(sep)

  total = sum(
    len(string) + sep_size
    for string in strings
  )

  total -= sep_size
  return Ok(total)


@_output_items
@_use_metadata(str.count)
def count(sub: str, *args: Args) -> Result[int]:
  if sub is None or sub == EMPTY_STR:
    return IntError

  strings, _ = _get_strings_sep(args)

  total = sum(
    string.count(sub)
    for string in strings
  )

  if not total:
    return NotFound(total)

  return Ok(total)


@_output_items
@_use_metadata(str.index)
def index(
  sub: str,
  *args: Args,
  start: int | None = None,
  end: int | None = None,
) -> Result[int]:
  if sub is None or sub == EMPTY_STR:
    return IntError

  strings, sep = _get_strings_sep(args)
  match_index: int | None
  index: int = START_INDEX

  for string in strings:
    line = f'{string}{sep}'

    try:
      match_index = line.index(sub, start, end)

    except ValueError as e:
      logging.debug(f"Caught ValueError({e})")
      match_index = None

    if match_index is not None:
      index += match_index
      return Ok(index)

    index += len(line)

  return NotFound()


@_output_items
@_use_metadata(str.rindex)
def rindex(
  sub: str,
  *args: Args,
  start: int | None = None,
  end: int | None = None,
) -> Result[int]:
  if sub is None or sub == EMPTY_STR:
    return IntError

  strings, sep = _get_strings_sep(args)
  match_index: int | None
  index: int = START_INDEX

  for string in strings:
    line = f'{string}{sep}'

    try:
      match_index = line.rindex(sub)

    except ValueError:
      match_index = None

    if match_index is not None:
      index += match_index
      return Ok(index)

    index += len(line)

  return NotFound()


@_output_items
@_use_metadata(str.find)
def find(
  sub: str | None = None,
  *args: Args,
  start: int | None = None,
  end: int | None = None,
) -> Result[int]:
  if sub is None or sub == EMPTY_STR:
    return IntError

  strings, sep = _get_strings_sep(args)
  match_index: int | None
  index: int = START_INDEX

  for string in strings:
    line = f'{string}{sep}'
    match_index = line.find(sub, start, end)

    if match_index is not NOT_FOUND:
      index += match_index
      return Ok(index)

    index += len(line)

  return NotFound(NOT_FOUND)


@_output_items
@_use_metadata(str.rfind)
def rfind(
  sub: str | None = None,
  *args: Args,
  start: int | None = None,
  end: int | None = None,
) -> Result[int]:
  if sub is None or sub == EMPTY_STR:
    return IntError

  strings, sep = _get_strings_sep(args)
  match_index: int | None
  index: int = START_INDEX

  for string in strings:
    line = f'{string}{sep}'
    match_index = line.rfind(sub, start, end)

    if match_index is not NOT_FOUND:
      index += match_index
      return Ok(index)

    index += len(line)

  return NotFound(NOT_FOUND)


__all__ = [
  'count',
  'find',
  'index',
  'length',
  'rfind',
  'rindex',
]



from setuptools import setup, find_packages
from pathlib import Path


requirements: list[str] = \
  Path('requirements.txt') \
    .read_text() \
    .splitlines()

pkgs: list[str] = find_packages(
  # include=[
  #  'strs',
  #  'strs.*',
  # ]
)


setup(
  name="strs",
  version="0.2.1",
  description="🧵 Easy string tools for the shell",
  url="https://github.com/alexdelorenzo/strs",
  author="Alex DeLorenzo",
  license="AGPL 3.0",
  packages=pkgs,
  install_requires=requirements,
  python_requires='>=3.10',
  entry_points={
    "console_scripts": [
      "strs = strs.__main__:_main",
      "str = strs.__main__:_main",
    ]
  }
)



from __future__ import annotations
from typing import Final

from .base import *
from .constants import *
from .decorators import *
from .input import *
from .process import *
from .types import *


NAME: Final[str] = 'strs'



from __future__ import annotations
from functools import partial

from ...core.constants import SPACE, ALL, FIRST, NEW_LINE, EMPTY_STR, SAME_LINE, NO_CMD_ERR
from ...core.decorators import _wrap_parse_print, _use_metadata
from ...core.process import _output_items
from ...core.input import _get_strings_sep
from ...core.types import Args, Items, StrSep, Result, Ok


upper = _wrap_parse_print(str.upper)
lower = _wrap_parse_print(str.lower)
capitalize = _wrap_parse_print(str.capitalize)
casefold = _wrap_parse_print(str.casefold)
swapcase = _wrap_parse_print(str.swapcase)
title = _wrap_parse_print(str.title)


@_output_items
@_use_metadata(str.center)
def center(
  width: int,
  *args: Args,
  fillchar: str = SPACE,
) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.center(width, fillchar)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.replace)
def replace(
  old: str,
  new: str,
  *args: Args,
  count: int = ALL
) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.replace(old, new, count)
    yield StrSep(string, sep)


replace_first = _use_metadata(replace)(
  partial(replace, count=FIRST)
)


@_output_items
@_use_metadata(str.split)
def split(on: str = NEW_LINE, *args: Args) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  if not sep:
    sep = NEW_LINE

  for string in strings:
    string = string.rstrip(NEW_LINE)
    split_strs = string.split(sep=on)

    for split_str in split_strs:
      if not split_str:
        continue

      yield StrSep(split_str, sep)


@_output_items
@_use_metadata(str.rsplit)
def rsplit(on: str = NEW_LINE, *args: Args) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    split_strs = string.rsplit(sep=on)

    for split_str in split_strs:
      if not split_str:
        continue

      yield StrSep(split_str, sep)


@_output_items
@_use_metadata(str.join)
def join(
  on: str = EMPTY_STR,
  *args: Args
) -> Items[StrSep | Result[str]]:
  strings, _ = _get_strings_sep(args)
  sep = SAME_LINE

  if first := next(strings, None):
    first = first.rstrip(NEW_LINE)

  elif first is None:
    return

  yield StrSep(first, sep)

  for string in strings:
    string = string.rstrip(NEW_LINE)
    yield StrSep(f'{on}{string}', sep)

  yield Ok(EMPTY_STR)


@_output_items
@_use_metadata(str.strip)
def strip(chars: str | None = None, *args: Args) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.strip(chars)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.lstrip)
def lstrip(chars: str | None = None, *args: Args) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.lstrip(chars)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.rstrip)
def rstrip(chars: str | None = None, *args: Args) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.rstrip(chars)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.expandtabs)
def expandtabs(*args: Args, tabsize: int = 8) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.expandtabs(tabsize)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.ljust)
def ljust(width: int, *args: Args, fillchar: str = SPACE) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.ljust(width, fillchar)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.rjust)
def rjust(width: int, *args: Args, fillchar: str = SPACE) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.rjust(width, fillchar)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.zfill)
def zfill(width: int, *args: Args) -> Items[StrSep]:
  strings, sep = _get_strings_sep(args)

  for string in strings:
    string = string.zfill(width)
    yield StrSep(string, sep)


@_output_items
@_use_metadata(str.partition)
def partition(sep: str, *args: Args) -> Items[StrSep]:
  strings, _ = _get_strings_sep(args)
  parts: tuple[str, str, str]

  for string in strings:
    parts = string.partition(sep)
    output = NEW_LINE.join(parts)
    yield StrSep(output)


@_output_items
@_use_metadata(str.rpartition)
def rpartition(sep: str, *args: Args) -> Items[StrSep]:
  strings, _ = _get_strings_sep(args)
  parts: tuple[str, str, str]

  for string in strings:
    parts = string.rpartition(sep)
    output = NEW_LINE.join(parts)
    yield StrSep(output)


@_use_metadata(str.format)
def format(fmt: str, *args: Args, **kwargs):
  strings, sep = _get_strings_sep(args)

  for string in strings:
    result = string.format(fmt, **kwargs)
    print(result, sep)

  raise NotImplementedError(NO_CMD_ERR)


@_use_metadata(str.format_map)
def format_map(**kwargs):
  raise NotImplementedError(NO_CMD_ERR)


__all__ = [
  'capitalize',
  'casefold',
  'center',
  'expandtabs',
  'format',
  'format_map',
  'join',
  'ljust',
  'lower',
  'lstrip',
  'partition',
  'replace',
  'rjust',
  'rpartition',
  'rsplit',
  'rstrip',
  'split',
  'strip',
  'swapcase',
  'title',
  'upper',
  'zfill',
]
